           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_1 = Csp_Prover:

(*********************************************************
              data type passed on channels
 *********************************************************)

typedecl init_data
typedecl request_data
typedecl response_data
typedecl exit_data

datatype Data = Init    init_data    | Exit     exit_data
              | Request request_data | Response response_data

(*********************************************************
                     event (channel)
 *********************************************************)

datatype Event = c Data

(*********************************************************
         abstract component description level
 *********************************************************)

datatype ACName = Acquirer | AcConfigManagement 
                | Terminal | TerminalConfigManagement

consts 
  ACDef :: "(ACName, Event) procDef"

primrec
  "ACDef (Acquirer) 
         = c ? x:(range Init) -> <AcConfigManagement>"

  "ACDef (AcConfigManagement)
         = c !! exit:(range Exit) -> SKIP
           |~| 
           c !! request:(range Request)
             -> c ? response:(range Response)
             -> <AcConfigManagement>"

  "ACDef (Terminal)
         = c !! init:(range Init) -> <TerminalConfigManagement>"

  "ACDef (TerminalConfigManagement)
         = c ? x -> IF (x:range Request) 
                    THEN c !! response:(range Response)
                          -> <TerminalConfigManagement>
                    ELSE IF (x:range Exit)
                         THEN SKIP
                         ELSE STOP"

consts
  AC :: "(ACName, Event) procRec"

defs
  AC_def:
   "AC == 
     LET ACDef
     IN (<Acquirer> |[range c]| <Terminal>)"

(*---------------------------------------------------------------*
 |                          NOTE                                 |
 |                                                               |
 | c ! v -> P       : sends a value v to c, then behaves like P. |
 |                                                               |
 | c ? x:X -> P(x)  : receives a value v from c if c in X,       |
 |                    then behaves like P(v).                    |
 |                                                               |
 | c !! x:X -> P(x) : sends a value v selected from X to c,      |
 |                    then behaves like P(v).                    |
 |                                                               |
 *---------------------------------------------------------------*)

(*********************************************************
                  architectual level
 *********************************************************)

datatype ArcName = Abstract | Loop

consts 
  ArcDef :: "(ArcName, Event) procDef"

primrec
  "ArcDef (Abstract)
          = c !! init:(range Init) -> <Loop>"

  "ArcDef (Loop)
          = c !! exit:(range Exit) -> SKIP
            |~| 
            c !! request:(range Request)
              -> c !! response:(range Response)
              -> <Loop>"

consts
  Arc :: "(ArcName, Event) procRec"

defs
  Arc_def:
   "Arc == 
     LET ArcDef IN <Abstract>"

(*********************************************************
        relating function between ACName and ArcName
 *********************************************************)

consts 
  Arc_to_AC :: "ArcName => (ACName, Event) proc"

primrec
  "Arc_to_AC (Abstract)
          = (<Acquirer> |[range c]| <Terminal>)"

  "Arc_to_AC (Loop)
          = (<AcConfigManagement> |[range c]| <TerminalConfigManagement>)"

(*********************************************************
        nohide and guarded lemmas (routine work)
 *********************************************************)

(*** To automatically unfold syntactic sugar ***)
declare csp_ss_def[simp]

lemma guardACDef[simp]: "!!C. guard ((ACDef) C)"
apply (induct_tac C, simp_all)
done

lemma nohideACDef[simp]: "!!C. nohide ((ACDef) C)"
apply (induct_tac C, simp_all)
done

lemma guardArcDef[simp]: "!!C. guard ((ArcDef) C)"
apply (induct_tac C, simp_all)
done

lemma nohideArcDef[simp]: "!!C. nohide ((ArcDef) C)"
apply (induct_tac C, simp_all)
done

declare csp_ss_def [simp del]

(*********************************************************
           a theorem for verifying Arc =F AC
 *********************************************************)

theorem ep2: "Arc =F AC"
apply (rule csp_fp_induct[of _ _ "Arc_to_AC"])
apply (rule)
apply (simp_all add: Arc_def AC_def)
apply (induct_tac C)

apply ( simp add: csp_ss_def image_iff | (rule csp_decompo) 
      | tactic {* csp_hnf_tac 1 *} | auto )+
done

(*--------------------------------------------------------*
 |                                                        |
 |        another proof of Arc =F AC (more manual)        |
 |                                                        |
 |           The proof is divided to two lemmas.          |
 |                                                        |
 *--------------------------------------------------------*)

(*** initializing ***)

lemma ep2_step1:
   "LET ACDef IN Rewrite (ArcDef Abstract) By Arc_to_AC =F 
    LET ACDef IN Arc_to_AC Abstract"
apply (tactic {* csp_unwind_tac 1 *})
apply (unfold csp_ss_def, tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo, simp)

apply (tactic {* csp_step_tac 1 *})
apply (auto)
apply (tactic {* csp_simp_tac 1 *})
done

(*** main loop ***)

lemma ep2_step2:
   "LET ACDef IN Rewrite (ArcDef Loop) By Arc_to_AC =F 
    LET ACDef IN Arc_to_AC Loop"
apply (tactic {* csp_unwind_tac 1 *})
apply (unfold csp_ss_def, tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo)

(* case of eixt *)
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_simp_tac 1 *})

(* case of loop *)

 (* request phase *)
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_simp_tac 1 *})

 (* response phase *)
apply (tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_simp_tac 1 *})
done

(**********************
       Arc =F AC ?
 **********************)

theorem ep2_man: "Arc =F AC"
  (***** by fixed point induction *****)
apply (rule csp_fp_induct[of _ _ "Arc_to_AC"])
apply (rule)

  (***** check guarded and no hiding operators *****)
apply (simp_all add: Arc_def AC_def)
apply (induct_tac C)
apply (simp add: ep2_step1 del: ArcDef.simps Arc_to_AC.simps)
apply (simp add: ep2_step2 del: ArcDef.simps Arc_to_AC.simps)
done

end
